@@ -51,8 +51,7 @@
 
 #include "xf86drm.h"
 #include "xf86drmMode.h"
-#include "intel_bufmgr.h"
-#include "i915_drm.h"
+#include "libkms.h"
 
 #ifdef HAVE_CAIRO
 #include <math.h>
@@ -111,6 +110,8 @@ struct type_name connector_type_names[] = {
 	{ DRM_MODE_CONNECTOR_DisplayPort, "displayport" },
 	{ DRM_MODE_CONNECTOR_HDMIA, "HDMI-A" },
 	{ DRM_MODE_CONNECTOR_HDMIB, "HDMI-B" },
+	{ DRM_MODE_CONNECTOR_TV, "TV" },
+	{ DRM_MODE_CONNECTOR_eDP, "embedded displayport" },
 };
 
 type_name_fn(connector_type)
@@ -352,37 +353,60 @@ connector_find_mode(struct connector *c)
 		c->crtc = c->encoder->crtc_id;
 }
 
-#ifdef HAVE_CAIRO
-
-static int
-create_test_buffer(drm_intel_bufmgr *bufmgr,
-		   int width, int height, int *stride_out, drm_intel_bo **bo_out)
+static struct kms_bo *
+allocate_buffer(struct kms_driver *kms,
+		int width, int height, int *stride)
 {
-	drm_intel_bo *bo;
-	unsigned int *fb_ptr;
-	int size, i, stride;
-	div_t d;
-	cairo_surface_t *surface;
-	cairo_t *cr;
-	char buf[64];
-	int x, y;
+	struct kms_bo *bo;
+	unsigned bo_attribs[] = {
+		KMS_WIDTH,   0,
+		KMS_HEIGHT,  0,
+		KMS_BO_TYPE, KMS_BO_TYPE_SCANOUT_X8R8G8B8,
+		KMS_TERMINATE_PROP_LIST
+	};
+	int ret;
 
-	surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
-	stride = cairo_image_surface_get_stride(surface);
-	size = stride * height;
-	fb_ptr = (unsigned int *) cairo_image_surface_get_data(surface);
+	bo_attribs[1] = width;
+	bo_attribs[3] = height;
 
-	/* paint the buffer with colored tiles */
-	for (i = 0; i < width * height; i++) {
-		d = div(i, width);
-		fb_ptr[i] = 0x00130502 * (d.quot >> 6) + 0x000a1120 * (d.rem >> 6);
+	ret = kms_bo_create(kms, bo_attribs, &bo);
+	if (ret) {
+		fprintf(stderr, "failed to alloc buffer: %s\n",
+			strerror(-ret));
+		return NULL;
 	}
 
+	ret = kms_bo_get_prop(bo, KMS_PITCH, stride);
+	if (ret) {
+		fprintf(stderr, "failed to retreive buffer stride: %s\n",
+			strerror(-ret));
+		kms_bo_destroy(&bo);
+		return NULL;
+	}
+
+	return bo;
+}
+
+static void
+make_pwetty(void *data, int width, int height, int stride)
+{
+#ifdef HAVE_CAIRO
+	cairo_surface_t *surface;
+	cairo_t *cr;
+	int x, y;
+
+	surface = cairo_image_surface_create_for_data(data,
+						      CAIRO_FORMAT_ARGB32,
+						      width, height,
+						      stride);
 	cr = cairo_create(surface);
+	cairo_surface_destroy(surface);
+
 	cairo_set_line_cap(cr, CAIRO_LINE_CAP_SQUARE);
 	for (x = 0; x < width; x += 250)
 		for (y = 0; y < height; y += 250) {
-			cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
+			char buf[64];
+
 			cairo_move_to(cr, x, y - 20);
 			cairo_line_to(cr, x, y + 20);
 			cairo_move_to(cr, x - 20, y);
@@ -395,6 +419,7 @@ create_test_buffer(drm_intel_bufmgr *bufmgr,
 			cairo_set_source_rgb(cr, 1, 1, 1);
 			cairo_set_line_width(cr, 2);
 			cairo_stroke(cr);
+
 			snprintf(buf, sizeof buf, "%d, %d", x, y);
 			cairo_move_to(cr, x + 20, y + 20);
 			cairo_text_path(cr, buf);
@@ -405,97 +430,74 @@ create_test_buffer(drm_intel_bufmgr *bufmgr,
 		}
 
 	cairo_destroy(cr);
-
-	bo = drm_intel_bo_alloc(bufmgr, "frontbuffer", size, 4096);
-	if (!bo) {
-		fprintf(stderr, "failed to alloc buffer: %s\n",
-			strerror(errno));
-		return -1;
-	}
-
-	drm_intel_bo_subdata(bo, 0, size, fb_ptr);
-
-	cairo_surface_destroy(surface);
-
-	*bo_out = bo;
-	*stride_out = stride;
-
-	return 0;
+#endif
 }
 
-#else
-
 static int
-create_test_buffer(drm_intel_bufmgr *bufmgr,
-		   int width, int height, int *stride_out, drm_intel_bo **bo_out)
+create_test_buffer(struct kms_driver *kms,
+		   int width, int height, int *stride_out,
+		   struct kms_bo **bo_out)
 {
-	drm_intel_bo *bo;
-	unsigned int *fb_ptr;
-	int size, ret, i, stride;
-	div_t d;
+	struct kms_bo *bo;
+	int ret, i, j, stride;
+	void *virtual;
 
-	/* Mode size at 32 bpp */
-	stride = width * 4;
-	size = stride * height;
-
-	bo = drm_intel_bo_alloc(bufmgr, "frontbuffer", size, 4096);
-	if (!bo) {
-		fprintf(stderr, "failed to alloc buffer: %s\n",
-			strerror(errno));
+	bo = allocate_buffer(kms, width, height, &stride);
+	if (!bo)
 		return -1;
-	}
 
-	ret = drm_intel_gem_bo_map_gtt(bo);
+	ret = kms_bo_map(bo, &virtual);
 	if (ret) {
-		fprintf(stderr, "failed to GTT map buffer: %s\n",
-			strerror(errno));
+		fprintf(stderr, "failed to map buffer: %s\n",
+			strerror(-ret));
+		kms_bo_destroy(&bo);
 		return -1;
 	}
 
-	fb_ptr = bo->virtual;
-
 	/* paint the buffer with colored tiles */
-	for (i = 0; i < width * height; i++) {
-		d = div(i, width);
-		fb_ptr[i] = 0x00130502 * (d.quot >> 6) + 0x000a1120 * (d.rem >> 6);
+	for (j = 0; j < height; j++) {
+		uint32_t *fb_ptr = (uint32_t*)((char*)virtual + j * stride);
+		for (i = 0; i < width; i++) {
+			div_t d = div(i, width);
+			fb_ptr[i] =
+				0x00130502 * (d.quot >> 6) +
+				0x000a1120 * (d.rem >> 6);
+		}
 	}
-	drm_intel_gem_bo_unmap_gtt(bo);
+
+	make_pwetty(virtual, width, height, stride);
+
+	kms_bo_unmap(bo);
 
 	*bo_out = bo;
 	*stride_out = stride;
-
 	return 0;
 }
 
-#endif
-
 static int
-create_grey_buffer(drm_intel_bufmgr *bufmgr,
-		   int width, int height, int *stride_out, drm_intel_bo **bo_out)
+create_grey_buffer(struct kms_driver *kms,
+		   int width, int height, int *stride_out,
+		   struct kms_bo **bo_out)
 {
-	drm_intel_bo *bo;
+	struct kms_bo *bo;
 	int size, ret, stride;
+	void *virtual;
 
-	/* Mode size at 32 bpp */
-	stride = width * 4;
-	size = stride * height;
-
-	bo = drm_intel_bo_alloc(bufmgr, "frontbuffer", size, 4096);
-	if (!bo) {
-		fprintf(stderr, "failed to alloc buffer: %s\n",
-			strerror(errno));
+	bo = allocate_buffer(kms, width, height, &stride);
+	if (!bo)
 		return -1;
-	}
 
-	ret = drm_intel_gem_bo_map_gtt(bo);
+	ret = kms_bo_map(bo, &virtual);
 	if (ret) {
-		fprintf(stderr, "failed to GTT map buffer: %s\n",
-			strerror(errno));
+		fprintf(stderr, "failed to map buffer: %s\n",
+			strerror(-ret));
+		kms_bo_destroy(&bo);
 		return -1;
 	}
 
-	memset(bo->virtual, 0x77, size);
-	drm_intel_gem_bo_unmap_gtt(bo);
+	size = stride * height;
+	memset(virtual, 0x77, size);
+	kms_bo_unmap(bo);
 
 	*bo_out = bo;
 	*stride_out = stride;
@@ -535,10 +537,11 @@ page_flip_handler(int fd, unsigned int frame,
 static void
 set_mode(struct connector *c, int count, int page_flip)
 {
-	drm_intel_bufmgr *bufmgr;
-	drm_intel_bo *bo, *other_bo;
+	struct kms_driver *kms;
+	struct kms_bo *bo, *other_bo;
 	unsigned int fb_id, other_fb_id;
 	int i, ret, width, height, x, stride;
+	unsigned handle;
 	drmEventContext evctx;
 
 	width = 0;
@@ -552,17 +555,18 @@ set_mode(struct connector *c, int count, int page_flip)
 			height = c[i].mode->vdisplay;
 	}
 
-	bufmgr = drm_intel_bufmgr_gem_init(fd, 2<<20);
-	if (!bufmgr) {
-		fprintf(stderr, "failed to init bufmgr: %s\n", strerror(errno));
+	ret = kms_create(fd, &kms);
+	if (ret) {
+		fprintf(stderr, "failed to create kms driver: %s\n",
+			strerror(-ret));
 		return;
 	}
 
-	if (create_test_buffer(bufmgr, width, height, &stride, &bo))
+	if (create_test_buffer(kms, width, height, &stride, &bo))
 		return;
 
-	ret = drmModeAddFB(fd, width, height, 32, 32, stride, bo->handle,
-			   &fb_id);
+	kms_bo_get_prop(bo, KMS_HANDLE, &handle);
+	ret = drmModeAddFB(fd, width, height, 32, 32, stride, handle, &fb_id);
 	if (ret) {
 		fprintf(stderr, "failed to add fb: %s\n", strerror(errno));
 		return;
@@ -588,11 +592,12 @@ set_mode(struct connector *c, int count, int page_flip)
 
 	if (!page_flip)
 		return;
-
-	if (create_grey_buffer(bufmgr, width, height, &stride, &other_bo))
+	
+	if (create_grey_buffer(kms, width, height, &stride, &other_bo))
 		return;
 
-	ret = drmModeAddFB(fd, width, height, 32, 32, stride, other_bo->handle,
+	kms_bo_get_prop(other_bo, KMS_HANDLE, &handle);
+	ret = drmModeAddFB(fd, width, height, 32, 32, stride, handle,
 			   &other_fb_id);
 	if (ret) {
 		fprintf(stderr, "failed to add fb: %s\n", strerror(errno));
@@ -609,7 +614,7 @@ set_mode(struct connector *c, int count, int page_flip)
 		c[i].swap_count = 0;
 		c[i].fb_id[0] = fb_id;
 		c[i].fb_id[1] = other_fb_id;
-		c[i].current_fb_id = fb_id;
+		c[i].current_fb_id = other_fb_id;
 	}
 
 	memset(&evctx, 0, sizeof evctx);
@@ -654,6 +659,10 @@ set_mode(struct connector *c, int count, int page_flip)
 
 		drmHandleEvent(fd, &evctx);
 	}
+
+	kms_bo_destroy(&bo);
+	kms_bo_destroy(&other_bo);
+	kms_destroy(&kms);
 }
 
 extern char *optarg;
@@ -679,6 +688,9 @@ void usage(char *name)
 
 static int page_flipping_supported(int fd)
 {
+	/*FIXME: generic ioctl needed? */
+	return 1;
+#if 0
 	int ret, value;
 	struct drm_i915_getparam gp;
 
@@ -692,6 +704,7 @@ static int page_flipping_supported(int fd)
 	}
 
 	return *gp.value;
+#endif
 }
 
 int main(int argc, char **argv)
